////////////////////////////////////////////////////////////////////////////////////////////////////
// PWM Software written by Chris Sykes.                                                           //
// NOTE: 16 bit Timers currently used. 32 Bit Timers need all shorts changed to longs!            //
// Software written and shared as is, no warrantys are given. Use at your own risk.               //
////////////////////////////////////////////////////////////////////////////////////////////////////
// Your MCU Clock Frequency: 16000000 or 16e6, retrieved by F_CPU.                                //
long fclock = F_CPU;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Most MCU's have:                                                                               //
//   0,                                                                                           //
//   8,                                                                                           //
//   64,                                                                                          //
//   256,                                                                                         //
//   1024                                                                                         //
// as Prescaler Values:                                                                           //
long Prescaler = 8;


////////////////////////////////////////////////////////////////////////////////////////////////////
// Serial Connection Hook:                                                                        //
bool Hook = true;


////////////////////////////////////////////////////////////////////////////////////////////////////
// The MCU Setup Method:                                                                          //
void setup() {//////////////////////////////////////////////////////////////////////////////////////

  // Init Serial:
  Serial.begin(115200);
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Configure Timer:                                                                               //
void ConfgureTimer(){///////////////////////////////////////////////////////////////////////////////

  // The counter simply overruns when it passes its maximum 16-bit value
  // (MAX = 0xFFFF (65535)) and then restarts from the BOTTOM (0x0000).
  
  // Timer One PWM Pin: 
  // Mega = Pin 12
  // Pro Micro = Pin 10
  // Check your Datasheet!
  pinMode(10, OUTPUT);
  
  // Reset Timer One:
  TCCR1A = 0;
  TCCR1B = 0;

  // ComA and B need to be set for Compare Match Pins:
  TCCR1A |= (1 << COM1A1) | (1 << COM1B1);
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Set Prescaler:                                                                                 //
void SetPrescaler(short prescaler){/////////////////////////////////////////////////////////////////

  ///////////////////////////////////////////////////////////////////////////////////////////////////////////
  // See your Datasheet! See: Clock Select Bit Description                                                 //
  //-------------------------------------------------------------------------------------------------------//
  // CS02     CS01     CS00     Description                                                                //
  //  0        0        0        No clock source (Timer/Counter stopped)                                   //
  //  0        0        1        clkI/O/(No prescaling)                                                    //
  //  0        1        0        clkI/O/8 (From prescaler)                                                 //
  //  0        1        1        clkI/O/64 (From prescaler)                                                //
  //  1        0        0        clkI/O/256 (From prescaler)                                               //
  //  1        0        1        clkI/O/1024 (From prescaler)                                              //
  //  1        1        0        External clock source on T0 pin. Clock on falling edge.                   //
  //  1        1        1        External clock source on T0 pin. Clock on rising edge.                    //
  ///////////////////////////////////////////////////////////////////////////////////////////////////////////

  // Configure Prescaler (CSX):
  switch (prescaler) {
    case 1:
      //  Set 0: clkI/O/(No prescaling)
      TCCR1B |= (1 << CS10);
      break;
    case 8:
      // Set 8: clkI/O/8 (From prescaler)
      TCCR1B |= (1 << CS11);
      break;
    case 64:
      // Set 64: clkI/O/64 (From prescaler)
      TCCR1B |= (1 << CS11) | (1 << CS10);
      break;
    case 256:
      // Set 256: clkI/O/256 (From prescaler)
      TCCR1B |= (1 << CS12);
      break;
    case 1024:
      // Set 1024: clkI/O/1024 (From prescaler)
      TCCR1B |= (1 << CS12) | (1 << CS10);
      break;
    default:
      // Set Off: No clock source (Timer/Counter stopped)
      TCCR1B |= (0 << CS12) | (0 << CS11) | (0 << CS10);
      break;
  }
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Set PWM Mode:                                                                                  //
void SetPWMMode(){//////////////////////////////////////////////////////////////////////////////////

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // See your Datasheet!                                                                                            //
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // Waveform Generation Mode Bit Description                                                                       //
  //----------------------------------------------------------------------------------------------------------------//
  //                                     Timer/Counter Mode of                    Update of       TOV Flag (1)(2)   //
  // Mode    WGM3   WGM2   WGM1   WGM0   Operation                     TOP        OCRx at         Set on            //
  //  0      0      0      0      0      Normal                        0xFFFF     Immediate       MAX               //
  //  1      0      0      0      1      PWM, Phase Correct, 8-bit     0x00FF     TOP             BOTTOM            //
  //  2      0      0      1      0      PWM, Phase Correct, 9-bit     0x01FF     TOP             BOTTOM            //
  //  3      0      0      1      1      PWM, Phase Correct, 10-bit    0x03FF     TOP             BOTTOM            //
  //  4      0      1      0      0      CTC                           OCRnA      Immediate       MAX               //
  //  5      0      1      0      1      Fast PWM, 8-bit               0x00FF     TOP             TOP               //
  //  6      0      1      1      0      Fast PWM, 9-bit               0x01FF     TOP             TOP               //
  //  7      0      1      1      1      Fast PWM, 10-bit              0x03FF     TOP             TOP               //
  //  8      1      0      0      0      PWM, PnFC                     ICRn       BOTTOM          BOTTOM            //
  //  9      1      0      0      1      PWM, PnFC                     OCRnA      BOTTOM          BOTTOM            //
  //  10     1      0      1      0      PWM, Phase Correct            ICRn       TOP             BOTTOM            //
  //  11     1      0      1      1      PWM, Phase Correct            OCRnA      TOP             BOTTOM            //
  //  12     1      1      0      0      CTC                           ICRn       Immediate       MAX               //
  //  13     1      1      0      1      (Reserved)                    –          –               –                 //
  //  14     1      1      1      0      Fast PWM                      ICRn       TOP             TOP               //
  //  15     1      1      1      1      Fast PWM                      OCRnA      TOP             TOP               //
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // NOTE: PnFC = Phase and Frequency Correct.

  // Fast PWM needs bits (WGM10 & WGM12) set to One:
  TCCR1A |= (1 << WGM11);

  // Configure Fast PWM Mode (WGM12 and 13):
  TCCR1B |= (1 << WGM12) | (1 << WGM13);
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// The MCU Loop Method:                                                                           //
void loop() {///////////////////////////////////////////////////////////////////////////////////////

  // Look for Serial Connection:
  if(Serial && Hook){

    // Connection Hook:
    Hook = false;
    
    // Print Welcome Message:
    Serial.println("Welcome to Serial PWM Controller by Chris Sykes!");
  }

  // Look for Serial Data:
  if(Serial.available()){

    // Read Serial String:
    String data = Serial.readString();

    // Check Data and Process: && data.endsWith(">") would be nice...
    if(data.length() > 0 && data.startsWith("<"))
      ProcessSerialData(data);

    // Flush Serial:
    Serial.flush();
  }
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Process the Serial Data:                                                                       //
void ProcessSerialData(String data){////////////////////////////////////////////////////////////////

  // Get Frequency: <15624|3906|1024> = 1Hz @ 25% Duty Cycle.
  long F = data.substring(1, data.indexOf('|')).toInt();
  long D = data.substring(data.indexOf('|') + 1, data.lastIndexOf('|')).toInt();
  Prescaler = data.substring(data.lastIndexOf('|') + 1, data.indexOf('>')).toInt();

  // Print Data:
  Serial.print("MCU set Frequency Register to: ");
  Serial.print(F);
  Serial.print(" or F: ");
  Serial.println(ConvertToFrequency(F));
  Serial.print("MCU set Duty Cycle Register to: ");
  Serial.println(D);
  Serial.print("MCU set Prescaler to: ");
  Serial.println(Prescaler);
  Serial.print("MCU Register Resolution: ");
  Serial.println(CalculateResolution(F));
  

  // Confgure Timer:
  ConfgureTimer();

  // Set the Prescaler:
  SetPrescaler(Prescaler);

  // Set PWM Mode:
  SetPWMMode();

  // Set Frequency:
  ICR1 = F;
  
  // Set Duty Cycle:
  OCR1B = D;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate the Frequency:                                                                       //
short ConvertToFrequency(long ocrx){////////////////////////////////////////////////////////////////

  // Calculate OCR1A from specified Frequency:
  return (short)(fclock / (Prescaler * (1 + ocrx)));
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate the Frequency:                                                                       //
short CalculateFrequency(double frequency){/////////////////////////////////////////////////////////

  // Calculate OCR1A from specified Frequency:
  return (short)(fclock / ((frequency * Prescaler) - 1));
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate the DutyCycle:                                                                       //
short CalculateDutyCycle(int period, double duty){//////////////////////////////////////////////////

  // Calculate OCR1B from OCR1A and specified Duty Cycle:
  return (short)(period * (duty / 100L));
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// Calculate Register Resolution:                                                                 //
double CalculateResolution(long ocrx){//////////////////////////////////////////////////////////////

  return (log((double)ocrx + 1.0D)/log(2.0D));
}
